

from tendril.utils.config import ConfigOption
from tendril.utils.config import ConfigOptionConstruct
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core',
           'tendril.config.auth']


class ServiceAuthParameters(ConfigOptionConstruct):
    @property
    def value(self):
        return {
            'provider': self.ctx["SERVICE_{}_AUTH_PROVIDER".format(self._parameters)],
            'realm': self.ctx["SERVICE_{}_AUTH_REALM".format(self._parameters)],
            'audience': self.ctx["SERVICE_{}_AUDIENCE".format(self._parameters)],
            'client_id': self.ctx["SERVICE_{}_CLIENT_ID".format(self._parameters)],
            'client_secret': self.ctx["SERVICE_{}_CLIENT_SECRET".format(self._parameters)],
            'ssl_verify': self.ctx["SERVICE_{}_SSL_VERIFICATION".format(self._parameters)],
        }


def _service_auth_config_template(service):
    return [
        ConfigOption(
            'SERVICE_{}_AUTH_PROVIDER'.format(service.upper()),
            "AUTH_PROVIDER",
            "Auth Provider which controls access to the {} service. Note that this is typically the same "
            "as the main auth provider, and use of multiple auth providers in a single instance or "
            "component is not presently tested or supported.".format(service)
        ),
        ConfigOption(
            'SERVICE_{}_AUTH_REALM'.format(service.upper()),
            f"AUTH_REALM",
            "Realm to for the {} service, if different from the Application's Realm".format(service)
        ),
        ConfigOption(
            'SERVICE_{}_AUDIENCE'.format(service.upper()),
            f"None",
            f"Audience to request in service access token for the {service} service."
            f" Expect Deprecation of this unused configuration item."
        ),
        ConfigOption(
            'SERVICE_{}_CLIENT_ID'.format(service.upper()),
            "None",
            f"Client ID to use for authenticating to the {service} service."
        ),
        ConfigOption(
            'SERVICE_{}_CLIENT_SECRET'.format(service.upper()),
            "None",
            f"Client Secret to use for authenticating to the {service} service.",
            masked=True
        ),
        ConfigOption(
            'SERVICE_{}_SSL_VERIFICATION'.format(service.upper()),
            "True",
            f"Whether SSL needs to be verified when authenticating to the {service} service."
        ),
        ServiceAuthParameters(
            'SERVICE_{}_AUTH_PARAMETERS'.format(service.upper()),
            service.upper(),
            f"Constructed Service Auth parameters instance for the "
            f"{service} service. This option is created by the code, and should not "
            f"be set directly in any config file."
        )
    ]


def load(manager):
    logger.debug("Loading {0}".format(__name__))
    config_elements_service_auth = []
    for service in manager.SERVICES:
        config_elements_service_auth += _service_auth_config_template(service)
    manager.load_elements(config_elements_service_auth,
                          doc="Authentication to Available Services")





from fastapi import APIRouter
from fastapi.responses import FileResponse
from fastapi import Depends

from tendril.apiserver.security import security
from tendril.apiserver.security import swagger_auth
from tendril.apiserver.security import AuthUserModel

from tendril.jwt.sign import generate_token_async

from tendril.config import AUTH_JWKS_PATH
from tendril.config import AUTH_JWT_API_ENABLED


jwt_services = APIRouter(prefix='/jwt',
                         tags=["JWT Services"])


@jwt_services.get("/jwks.json")
async def jwks():
    return FileResponse(AUTH_JWKS_PATH)


@jwt_services.get("/{domain}", dependencies=[Depends(swagger_auth)])
async def get_domain_token(domain: str, user: AuthUserModel = security()):
    token = await generate_token_async(user, domain)
    return {'jwt': token}


if AUTH_JWT_API_ENABLED:
    routers = [
        jwt_services
    ]
else:
    routers = []





from httpx import Auth
from tendril.authn import provider
from tendril import config

from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)


class IntramuralAuthenticator(Auth):
    requires_response_body = True

    def __init__(self, service):
        self._realm = None
        self._audience = None
        self._client_id = None
        self._client_secret = None
        self._get_service_auth_params(service)
        self._access_token = None

    def _get_service_auth_params(self, service):
        params = getattr(config, 'SERVICE_{}_AUTH_PARAMETERS'.format(service.upper()), None)
        if not params:
            raise ValueError("Service {} does not have authentication "
                             "parameters configured".format(service.upper()))
        if params['provider'] != provider.provider_name:
            raise ValueError("Service Authentication Provider must be the same as the application "
                             "authentication provider ")
        self._realm = params['realm']
        self._audience = params['audience']
        self._client_id = params['client_id']
        self._client_secret = params['client_secret']
        self._ssl_verify = params['ssl_verify']

    def sync_auth_flow(self, request):
        raise NotImplementedError

    async def async_auth_flow(self, request):
        if self._access_token:
            request.headers["Authorization"] = "Bearer " + self._access_token
            response = yield request

        if not self._access_token or response.status_code == 401:
            # If the server issues a 401 response, then issue a request to
            # refresh tokens, and resend the request.
            await self.async_get_access_token()
            request.headers["Authorization"] = "Bearer " + self._access_token
            yield request

    async def async_get_access_token(self):
        logger.info(f"Requesting intramural access token for {self._audience} from {provider.provider_name}")
        self._access_token = await provider.get_service_access_token(
            realm=self._realm,
            audience=self._audience,
            client_id=self._client_id,
            client_secret=self._client_secret
        )
        logger.info(f"Got intramural access token for {self._audience}")






from tendril.utils.fsutils import load_provider
from tendril.config import AUTHZ_PROVIDER
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)


try:
    AuthzProvider = load_provider(AUTHZ_PROVIDER, "tendril.authz.providers")
    provider_name = AUTHZ_PROVIDER
except ImportError as e:
    logger.critical(e)
    raise


def get_current_scopes():
    return AuthzProvider.get_current_scopes()


def commit_scopes(scopes):
    return AuthzProvider.commit_scopes(scopes)


def get_user_scopes(user_id):
    return AuthzProvider.get_user_scopes(user_id)


def add_user_scopes(user_id, scopes):
    return AuthzProvider.add_user_scopes(user_id, scopes)


def remove_user_scopes(user_id, scopes):
    return AuthzProvider.remove_user_scopes(user_id, scopes)





from tendril.authn.provider import AuthProvider
from tendril.authn.provider import provider_name
from tendril.authn.scaffold import wrap_security
from tendril.authn.scaffold import null_dependency

from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)

# This motif is based on Auth0's auth_dependency, where auth0's implicit scheme is provided
# for swagger integration. If a provider does not need this, it this can be replaced with a
# dummy function that just returns None.
swagger_auth = getattr(AuthProvider, "swagger_auth", null_dependency)
security = wrap_security(provider_name, AuthProvider.security)
AuthUserModel = AuthProvider.AuthUserModel





from fastapi import APIRouter
from fastapi import Depends
from fastapi import BackgroundTasks

from tendril.apiserver.security import security
from tendril.apiserver.security import swagger_auth
from tendril.apiserver.security import AuthUserModel

from tendril.authn.users import verify_user_registration
from tendril.authn.users import get_user_profile
from tendril.authn.users import get_user_stub


user_services = APIRouter(prefix='/user',
                          tags=["User Authentication Services"],
                          dependencies=[Depends(swagger_auth)])


@user_services.get("/verify")
async def verify(background_tasks: BackgroundTasks,
                 user: AuthUserModel = security()):
    verify_user_registration(user, background_tasks=background_tasks)
    return {"message": "Logged in User Verified."}


@user_services.get("/profile/me")
async def my_profile(user: AuthUserModel = security()):
    return await get_user_profile(user)


@user_services.get("/stub", dependencies=[security()])
async def user_stub(user_id: str):
    return await get_user_stub(user_id)


routers = [
    user_services
]





from tendril.config import AUTH_PROVIDER
from tendril.config import AUTH_INTRAMURAL_TOKEN_CACHING
from tendril.utils.fsutils import load_provider

from .db.controller import register_provider

from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)

try:
    AuthProvider = load_provider(AUTH_PROVIDER, "tendril.authn.providers")
    provider_name = AUTH_PROVIDER
except ImportError as e:
    logger.critical(e)
    raise


def _cache_key_func(audience, client_id, client_secret):
    return f"{provider_name}:{client_id}:{audience}"


if AUTH_INTRAMURAL_TOKEN_CACHING == 'platform':
    logger.info("Using platform level caching for Intramural Authentication Tokens")
    from tendril.caching import platform_cache as intramural_cache
else:
    logger.info("Not caching Intramural Authentication Tokens")
    from tendril.caching import no_cache as intramural_cache


@intramural_cache(namespace='intramuraltokens', ttl=3600, key=_cache_key_func)
async def get_service_access_token(realm, audience, client_id, client_secret, ssl_verify=True):
    """
    Generic API for intramural authenticators to request access tokens.
    Each provider module (Auth0, Keycloak, etc.) must implement this.
    """
    if not hasattr(AuthProvider, "get_service_access_token"):
        raise NotImplementedError(
            f"AuthProvider {provider_name} does not implement get_service_access_token"
        )

    return await AuthProvider.get_service_access_token(realm, audience, client_id, client_secret, ssl_verify)

def init():
    AuthProvider.init()
    register_provider(provider_name)

init()



#
#
# ### TODO DEPRECATE
#
# import string
# import secrets
#
# from tendril.config import AUTH_MECHANIZED_USER_DOMAIN
# from tendril.config import AUTH_MECHANIZED_PROVIDER
# from tendril.utils.fsutils import load_provider
# from .scaffold import null_dependency
# from .scaffold import wrap_security
#
# from tendril.utils import log
# logger = log.get_logger(__name__, log.DEBUG)
#
# enabled = False
#
# if not AUTH_MECHANIZED_PROVIDER:
#     logger.warning("AUTH_MECHANIZED_PROVIDER is not set. "
#                    "Support for Mechanized Users will not be available.")
#     MechanizedAuthProvider = None
#     mechanized_provider_name = None
#     swagger_auth_mechanized = null_dependency
#     security_mechanized = null_dependency
#     AuthMechanizedUserModel = None
#
# else:
#     try:
#         MechanizedAuthProvider = load_provider(AUTH_MECHANIZED_PROVIDER, "tendril.authn.providers")
#         mechanized_provider_name = AUTH_MECHANIZED_PROVIDER
#         if MechanizedAuthProvider.supports_mechanized:
#             enabled = True
#         else:
#             logger.critical(f"The configured AUTH_MECHANIZED_PROVIDER '{AUTH_MECHANIZED_PROVIDER}' "
#                             f"does not support mechanized users. Support for Mechanized Users "
#                             f"will not be available.")
#     except ImportError as e:
#         logger.critical(e)
#         raise
#
#     swagger_auth_mechanized = getattr(MechanizedAuthProvider, "swagger_auth", null_dependency)
#     security_mechanized = wrap_security(mechanized_provider_name, MechanizedAuthProvider.security)
#     get_mechanized_provider_user_profile = MechanizedAuthProvider.get_mechanized_user_profile
#     AuthMechanizedUserModel = MechanizedAuthProvider.AuthUserModel
#
#
# def _generate_password(length=32):
#     alphabet = string.ascii_letters + string.digits
#     password = ''.join(secrets.choice(alphabet) for i in range(length))
#     return password
#
#
# def _get_mechanized_user_username(username, prefix):
#     return f'{prefix}-{username}'
#
#
# def _get_mechanized_user_email(username, prefix):
#     return f'{username}@{prefix}.{AUTH_MECHANIZED_USER_DOMAIN}'
#
#
# def _get_mechanized_user_name(username, prefix):
#     return f'{prefix} {username}'
#
#
# def create_mechanized_user(username, prefix, role=None, password=None):
#     if not enabled:
#         raise RuntimeError("Mechanized users are not supported. There is probably no mechanized auth provider configured.")
#     if not password:
#         password = _generate_password()
#     MechanizedAuthProvider.create_user(
#         email=_get_mechanized_user_email(username, prefix),
#         username=_get_mechanized_user_username(username, prefix),
#         name=_get_mechanized_user_name(username, prefix),
#         role=role,
#         password=password, mechanized=True
#     )
#     return password
#
#
# def find_mechanized_user_by_email(email):
#     if not enabled:
#         raise RuntimeError("Mechanized users are not supported. There is probably no mechanized auth provider configured.")
#     return MechanizedAuthProvider.find_user_by_email(email, mechanized=True)
#
#
# def set_mechanized_user_password(user_id, password=None):
#     if not enabled:
#         raise RuntimeError("Mechanized users are not supported. There is probably no mechanized auth provider configured.")
#     if not password:
#         password = _generate_password()
#     MechanizedAuthProvider.set_user_password(user_id, password, mechanized=True)
#     return password





import inspect
from tendril.authz import domains
from tendril.authn.provider import AuthProvider
from tendril.authn.provider import provider_name

from .db.model import User
from .db.controller import register_user
from .db.controller import get_user_by_id

from tendril.utils import log
logger = log.get_logger(__name__, log.DEBUG)

AuthUserModel = AuthProvider.AuthUserModel
get_provider_user_profile = AuthProvider.get_user_profile


def preprocess_user(user):
    if isinstance(user, AuthUserModel):
        user = user.id
    elif isinstance(user, User):
        user = user.id
    elif isinstance(user, int):
        user = get_user_by_id(user).puid
    return user


async def get_user_profile(user):
    user = preprocess_user(user)
    profile = {}

    result = get_provider_user_profile(user)
    if inspect.isawaitable(result):
        result = await result
    profile[provider_name] = result
    return profile


async def expand_user_stub(v, **kwargs):
    if inspect.isawaitable(v):
        v = await v
    if isinstance(v, str):
        return await get_user_stub(v)
    return v


async def get_user_stub(user):
    user = preprocess_user(user)
    result = AuthProvider.get_user_stub(user)
    if inspect.isawaitable(result):
        result = await result
    return result


async def get_user_email(user):
    profile = await get_user_profile(user)
    try:
        return profile[provider_name]['email']
    except KeyError:
        raise AttributeError(f"Could not find an email for user {user}")


def verify_user_registration(user, background_tasks=None):
    user_id = preprocess_user(user)
    user, first_login = register_user(user_id, provider_name)
    if background_tasks:
        background_tasks.add_task(domains.upsert,
                                  user=user,
                                  first_login=first_login)



from tendril.utils.config import ConfigOption
from tendril.utils import log
logger = log.get_logger(__name__, log.DEFAULT)

depends = ['tendril.config.core']


config_elements_app = [
    ConfigOption(
        'AUTH_ENABLE_LOGIN',
        "True",
        "Whether to enable login on this application. The type of this login depends "
        "on the configuration of the underlying auth provider. This is usually going to be "
        "True."
    ),
    ConfigOption(
        'AUTH_ENABLE_ADMIN',
        "True",
        "Whether user administration primitives are available on this application. "
        "This is currently required because Admin privileges are generally required for "
        "User Profile retrieval, and the ability to retrieve the user profile is "
        "presently assumed."
    ),
    ConfigOption(
        'AUTH_ENABLE_ADMIN_REALMS',
        "False",
        "Whether to Enable administration of other realms / clients / connections "
        "from this application. If true, ensure that the remote realms / clients / "
        "connections for the underlying auth providers are correctly configured. This "
        "is current generally used on Actual to manage device registration for iotedge."
    ),
    ConfigOption(
        'AUTH_REALMS',
        "[]",
        "Realms for which administration should be allowed from this application. "
        "This should only be configured on applications / components from which external "
        "realms need to be configured. Note that this presently is to be implemented in "
        "the underlying provider. The realm for the current application generally need not "
        "be included here, and the admin client for this realm is available with realm=None."
    ),
    ConfigOption(
        'SERVICES',
        "[]",
        "Services accessible to the application. These services are typically "
        "expected to be authenticated to with the client credentials grant. The necessary "
        "credentials may be configured for each service listed here as per the "
        "auth_services config construct.",
    )
]

config_elements_auth = [
    ConfigOption(
        'AUTH_PROVIDER',
        "'auth0'",
        "AuthN Provider"
    ),
    ConfigOption(
        'AUTHZ_PROVIDER',
        "AUTH_PROVIDER",
        "AuthZ Provider"
    ),
    ConfigOption(
        'AUTH_REALM',
        "None",
        "Realm to use for authentication. Required in multi-realm applications and when using "
        "providers which require realm specification such as keycloak."
    ),
    ConfigOption(
        "AUTH_AUDIENCE_BASE",
        "''",
        "Prefix for Audiences"
    ),
    ConfigOption(
        "AUTH_ISSUER_BASE",
        "'https://tendril.link/auth/'",
        "Prefix for Issuers"
    ),
    ConfigOption(
        "AUTH_COMPONENT",
        "COMPONENT_NAME",
        "Component for Audiences and Issuers"
    ),
    ConfigOption(
        "AUTH_AUDIENCE",
        "AUTH_AUDIENCE_BASE + AUTH_COMPONENT",
        "Audience required by client JWTs to interact with this component"
    ),
    ConfigOption(
        "AUTH_ISSUER",
        "AUTH_ISSUER_BASE + AUTH_COMPONENT",
        "Issuer specified in JWTs generated by this component"
    ),
    ConfigOption(
        'AUTH_PATCH_SCOPES_ON_STARTUP',
        'False',
        "Whether to patch scopes on the AuthZ provider at startup."
    ),
    ConfigOption(
        'AUTH_USERINFO_CACHING',
        "None",
        "Whether to cache userinfo acquired from the auth provider. "
        "Set to 'platform' for using platform-level caching, using redis. "
        "Other options not presently implemented."
    ),
    ConfigOption(
        'AUTH_INTRAMURAL_TOKEN_CACHING',
        "None",
        "Whether to cache intramural access tokens. This should generally "
        "be disabled for applications where security is a serious concern. "
        "Set to 'platform' for using platform-level caching, using redis. "
        "Other options not presently implemented."
    )
]

config_elements_jwt = [
    ConfigOption(
        'AUTH_JWT_API_ENABLED',
        "True",
        "Whether to enable JWT generation API"
    ),
    ConfigOption(
        'AUTH_JWKS_PATH',
        'os.path.join(INSTANCE_ROOT, "jwt", "jwks.json")',
        "Path to the jwks.json file to serve."
    ),
    ConfigOption(
        'AUTH_JWT_PRIVATE_KEY',
        'os.path.join(INSTANCE_ROOT, "jwt", "privateKey.pem")',
        "Path to the private key to use to sign JWTs."
    ),
    ConfigOption(
        'AUTH_JWT_PUBLIC_KEY',
        'os.path.join(INSTANCE_ROOT, "jwt", "publicKey.pem")',
        "Path to the public key to use to verify JWTs."
    ),
    ConfigOption(
        'AUTH_JWT_ISSUER',
        'AUTH_ISSUER',
        "Issuer to specify in the issued JWT claims."
    ),
    ConfigOption(
        'AUTH_JWT_TTL',
        '0',
        "TTL in seconds before the issued JWT can be used."
    ),
    ConfigOption(
        'AUTH_JWT_VALIDITY',
        '3600',
        "Duration in seconds before the issued JWT is expired."
    ),
    ConfigOption(
        'AUTH_JWT_ALGORITHMS',
        '["RS256", "ES256", "HS256"]',
        "Algorithm to use to sign or verify JWTs. The first element used for encode."
    )
]

# config_elements_mechanized = [
#     ConfigOption(
#         'AUTH_MECHANIZED_PROVIDER',
#         'AUTH_PROVIDER',
#         "Provider to use for mechanized users, if different."
#     ),
#     ConfigOption(
#         'AUTH_MECHANIZED_AUDIENCE',
#         'AUTH_AUDIENCE_BASE + iotedge',
#         "Audience to expect in the Mechanized Keycloak Tokens"
#     ),
#     ConfigOption(
#         'AUTH_MECHANIZED_USER_DOMAIN',
#         '"tendril.link"',
#         "Domain to use for mechanized user email addresses."
#     ),
# ]

def load(manager):
    logger.debug("Loading {0}".format(__name__))
    manager.load_elements(config_elements_app,
                          doc="Authentication Features Configuration",)
    manager.load_elements(config_elements_auth,
                          doc="Authentication Configuration")
    manager.load_elements(config_elements_jwt,
                          doc="JWT Infrastructure Configuration")
    # manager.load_elements(config_elements_mechanized,
    #                       doc="Mechanized Users Configuration")





from typing import Optional
from functools import lru_cache

"""
This module implements a caching layer around provider-specific `security` functions.

Factory Pattern:
----------------
Each provider defines its own `security(scopes=None)` returning (usually) a FastAPI
Security() dependency. The common layer wraps that function with caching, so that
identical scope combinations reuse the same Security() instance across endpoints and
modules.
"""

def wrap_security(provider_name: str, provider_security):
    """
    Wrap provider_security(scopes=None) with a cache that preserves
    FastAPI's function signature for OpenAPI introspection.
    """
    @lru_cache(maxsize=None)
    def _cached(scopes_key: str):
        scopes = scopes_key.split(" ") if scopes_key else []
        return provider_security(scopes=scopes)

    def security(scopes: Optional[list[str]] = None):
        scope_key = " ".join(sorted(scopes or []))
        return _cached(scope_key)

    return security


def null_dependency():
    return None
